@@ -232,6 +232,30 @@ typedef struct
   unsigned char fill_1[4];
 } Elf64_External_VMS_IMAGE_RELA;
 
+/* Note segments.  VMS is special as it uses 64-bit entries.  */
+
+typedef struct {
+  unsigned char	namesz[8];		/* Size of entry's owner string */
+  unsigned char	descsz[8];		/* Size of the note descriptor */
+  unsigned char	type[8];		/* Interpretation of the descriptor */
+  char		name[1];		/* Start of the name+desc data */
+} Elf64_External_VMS_Note;
+
+#define NT_VMS_MHD         1 /* Object module name, version, and date/time.  */
+#define NT_VMS_LNM         2 /* Language processor name.  */
+#define NT_VMS_SRC         3 /* Source files.  */
+#define NT_VMS_TITLE       4 /* Title text.  */
+#define NT_VMS_EIDC        5 /* Entity ident consistency check.  */
+#define NT_VMS_FPMODE      6 /* Whole program floating-point mode.  */
+#define NT_VMS_LINKTIME  101 /* Date/time image was linked.  */
+#define NT_VMS_IMGNAM    102 /* Image name string.  */
+#define NT_VMS_IMGID     103 /* Image ident string.  */
+#define NT_VMS_LINKID    104 /* Linker ident string.  */
+#define NT_VMS_IMGBID    105 /* Image build ident string.  */
+#define NT_VMS_GSTNAM    106 /* Global Symbol Table Name.  */
+#define NT_VMS_ORIG_DYN  107 /* Original setting of dynamic data.  */
+#define NT_VMS_PATCHTIME 108 /* Date/time of last patch.  */
+
 /* IA64-specific relocation types: */
 
 /* Relocs apply to specific instructions within a bundle.  The least
